package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.StringTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.block.BlockEnergyCounter;
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.menu.EnergyCounterMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.lang.ref.WeakReference;
import java.math.BigDecimal;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityEnergyCounter extends BlockEntityMattery {
	protected BigDecimal passed = BigDecimal.ZERO;

	public BigDecimal getPassed() {
		return passed;
	}

	public BlockEntityEnergyCounter(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.ENERGY_COUNTER, p_155229_, p_155230_);
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.putString("passed", passed.toString());
		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);

		if (nbt.get("passed") instanceof StringTag tag)
			passed = new BigDecimal(tag.getAsString());
	}

	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.energy_counter");

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new EnergyCounterMenu(containerID, inventory, this);
	}

	public final EnergyCounterCap input = new EnergyCounterCap(true);
	public final EnergyCounterCap output = new EnergyCounterCap(false);

	protected LazyOptional<IMatteryEnergyStorage> input_cap_mte = LazyOptional.empty();
	protected LazyOptional<IEnergyStorage> input_cap_fe = LazyOptional.empty();
	protected LazyOptional<IMatteryEnergyStorage> output_cap_mte = LazyOptional.empty();
	protected LazyOptional<IEnergyStorage> output_cap_fe = LazyOptional.empty();

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (p_155231_ instanceof ServerLevel level) {
			OverdriveThatMatters.tickOnce(level, this::checkSurroundings);
		}
	}

	public class EnergyCounterCap implements IMatteryEnergyStorage {
		public final boolean is_input;

		public EnergyCounterCap(boolean is_input) {
			this.is_input = is_input;
		}

		private EnergyCounterCap opposite() {
			return is_input ? output : input;
		}

		@Override
		public BigDecimal extractEnergyOuter(BigDecimal howMuch, boolean simulate) {
			return extractEnergyInner(howMuch, simulate);
		}

		@Override
		public BigDecimal extractEnergyInner(BigDecimal howMuch, boolean simulate) {
			if (is_input)
				return BigDecimal.ZERO;

			if (input_cap_mte.isPresent()) {
				final var value = input_cap_mte.resolve().get().extractEnergyOuter(howMuch, simulate);

				if (!simulate) {
					passed = passed.add(value);
				}

				return value;
			}

			if (input_cap_fe.isPresent()) {
				final var value = MatteryCapability.drainFE(input_cap_fe.resolve().get(), howMuch, simulate);

				if (!simulate) {
					passed = passed.add(value);
				}

				return value;
			}

			return BigDecimal.ZERO;
		}

		@Override
		public BigDecimal receiveEnergyOuter(BigDecimal howMuch, boolean simulate) {
			return receiveEnergyInner(howMuch, simulate);
		}

		@Override
		public BigDecimal receiveEnergyInner(BigDecimal howMuch, boolean simulate) {
			if (!is_input)
				return BigDecimal.ZERO;

			if (output_cap_mte.isPresent()) {
				final var value = output_cap_mte.resolve().get().receiveEnergyOuter(howMuch, simulate);

				if (!simulate) {
					passed = passed.add(value);
				}

				return value;
			}

			if (output_cap_fe.isPresent()) {
				final var value = MatteryCapability.floodFE(output_cap_fe.resolve().get(), howMuch, simulate);

				if (!simulate) {
					passed = passed.add(value);
				}

				return value;
			}

			return BigDecimal.ZERO;
		}

		@Override
		public BigDecimal getBatteryLevel() {
			if (is_input) {
				if (output_cap_mte.isPresent()) {
					return output_cap_mte.resolve().get().getBatteryLevel();
				}

				if (output_cap_fe.isPresent()) {
					return new BigDecimal(output_cap_fe.resolve().get().getEnergyStored());
				}
			} else {
				if (input_cap_mte.isPresent()) {
					return input_cap_mte.resolve().get().getBatteryLevel();
				}

				if (input_cap_fe.isPresent()) {
					return new BigDecimal(input_cap_fe.resolve().get().getEnergyStored());
				}
			}

			return BigDecimal.ZERO;
		}

		@Override
		public BigDecimal getMaxBatteryLevel() {
			if (is_input) {
				if (output_cap_mte.isPresent()) {
					return output_cap_mte.resolve().get().getMaxBatteryLevel();
				}

				if (output_cap_fe.isPresent()) {
					return new BigDecimal(output_cap_fe.resolve().get().getMaxEnergyStored());
				}
			} else {
				if (input_cap_mte.isPresent()) {
					return input_cap_mte.resolve().get().getMaxBatteryLevel();
				}

				if (input_cap_fe.isPresent()) {
					return new BigDecimal(input_cap_fe.resolve().get().getMaxEnergyStored());
				}
			}

			return BigDecimal.ZERO;
		}

		@Override
		public boolean canExtract() {
			return !is_input;
		}

		@Override
		public boolean canReceive() {
			return is_input;
		}
	}

	protected LazyOptional<IMatteryEnergyStorage> input_resolver = LazyOptional.of(() -> input);
	protected LazyOptional<IMatteryEnergyStorage> output_resolver = LazyOptional.of(() -> output);

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		input_resolver.invalidate();
		output_resolver.invalidate();
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		input_resolver = LazyOptional.of(() -> input);
		output_resolver = LazyOptional.of(() -> output);
	}

	@Override
	@SuppressWarnings("deprecation")
	public void setBlockState(BlockState p_155251_) {
		final var old = getBlockState();

		super.setBlockState(p_155251_);

		if (p_155251_ != old && p_155251_.getValue(BlockEnergyCounter.INPUT_DIRECTION) != old.getValue(BlockEnergyCounter.INPUT_DIRECTION)) {
			input_resolver.invalidate();
			output_resolver.invalidate();

			input_resolver = LazyOptional.of(() -> input);
			output_resolver = LazyOptional.of(() -> output);

			if (level != null)
				checkSurroundings(level);
		}
	}

	private <T> LazyOptional<T> getAndBind(Level level, LazyOptional<T> old, Capability<T> cap, Direction side) {
		final var ent = level.getBlockEntity(getBlockPos().offset(side.getNormal()));

		if (ent == null)
			return LazyOptional.empty();

		final var resolve = ent.getCapability(cap, side.getOpposite());

		if (resolve != old) {
			final var weak = new WeakReference<>(this);

			resolve.addListener((l) -> {
				final var get = weak.get();

				if (get != null && get.level != null) {
					get.checkSurroundings(get.level);
				}
			});

			return resolve;
		}

		return old;
	}

	public void checkSurroundings(Level level) {
		if (isRemoved() || !(level instanceof ServerLevel))
			return;

		input_cap_mte = getAndBind(level, input_cap_mte, MatteryCapability.ENERGY, getBlockState().getValue(BlockEnergyCounter.INPUT_DIRECTION));
		input_cap_fe = getAndBind(level, input_cap_fe, CapabilityEnergy.ENERGY, getBlockState().getValue(BlockEnergyCounter.INPUT_DIRECTION));

		output_cap_mte = getAndBind(level, output_cap_mte, MatteryCapability.ENERGY, getBlockState().getValue(BlockEnergyCounter.INPUT_DIRECTION).getOpposite());
		output_cap_fe = getAndBind(level, output_cap_fe, CapabilityEnergy.ENERGY, getBlockState().getValue(BlockEnergyCounter.INPUT_DIRECTION).getOpposite());
	}

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (side == null || isRemoved())
			return super.getCapability(cap, side);

		if (side == getBlockState().getValue(BlockEnergyCounter.INPUT_DIRECTION)) {
			if (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY) {
				return input_resolver.cast();
			}
		} else if (side == getBlockState().getValue(BlockEnergyCounter.INPUT_DIRECTION).getOpposite()) {
			if (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY) {
				return output_resolver.cast();
			}
		}

		return super.getCapability(cap, side);
	}
}
